@@ -1,36 +1,48 @@
-/* except.c -- exception mechanism */
+/* except.c -- exception mechanism ($Revision: 1.6 $) */
 
 #include "es.h"
 #include "print.h"
 
 /* globals */
-Handler childhandler = { NULL, NULL };
-Handler *bottomhandler = NULL;
+Handler *tophandler = NULL;
+Handler *roothandler = NULL;
 List *exception = NULL;
+Push *pushlist = NULL;
 
 /* pophandler -- remove a handler */
 extern void pophandler(Handler *handler) {
-	assert(bottomhandler == handler);
+	assert(tophandler == handler);
 	assert(handler->rootlist == rootlist);
-	bottomhandler = handler->up;
+	tophandler = handler->up;
 }
 
 /* throw -- raise an exception */
-extern noreturn throw(List *exc) {
-	Handler *handler = bottomhandler;
+extern noreturn throw(List *e) {
+	Handler *handler = tophandler;
 
-	assert(gcblocked == 0);
-	assert(exc != NULL);
+	assert(!gcisblocked());
+	assert(e != NULL);
 	assert(handler != NULL);
-	bottomhandler = handler->up;
+	tophandler = handler->up;
+	
+	while (pushlist != handler->pushlist) {
+		rootlist = &pushlist->defnroot;
+		varpop(pushlist);
+	}
+
+#if ASSERTIONS
 	for (; rootlist != handler->rootlist; rootlist = rootlist->next)
 		assert(rootlist != NULL);
-	exception = exc;
+#else
+	rootlist = handler->rootlist;
+#endif
+	exception = e;
 	longjmp(handler->label, 1);
+	NOTREACHED;
 }
 
 /* fail -- pass a user catchable error up the exception chain */
-extern noreturn fail VARARGS1(const char *, fmt) {
+extern noreturn fail VARARGS2(const char *, from, const char *, fmt) {
 	char *s;
 	va_list args;
 
@@ -40,14 +52,15 @@ extern noreturn fail VARARGS1(const char *, fmt) {
 
 	gcdisable(0);
 	Ref(List *, e, mklist(mkterm("error", NULL),
-			      mklist(mkterm(s, NULL), NULL)));
-	gcenable();
+			      mklist(mkterm((char *) from, NULL),
+				     mklist(mkterm(s, NULL), NULL))));
+	while (gcisblocked())
+		gcenable();
 	throw(e);
 	RefEnd(e);
 }
 
 /* newchildcatcher -- remove the current handler chain for a new child */
 extern void newchildcatcher(void) {
-	bottomhandler = &childhandler;
-	/* TODO: flush some of the rootlist? */
+	tophandler = roothandler;
 }
